function [sources,mixmat,sepmat,fvals]=cfica(signals,varargin)
% CFICA Complex-valued FastICA 
%
%  [sources,mixmat,sepmat]=cfica(signals,...)
%
%   signals: mixed signals. 3D array, channel x time x epoch
%   sources: separated sources. 3D array, channnel x time x epoch
%   mixmat : mixing matrix. #signals x #sources
%   sepmat : separation (demixing) matrix.
%     * sources(:,:) = sepmat*signals(:,:)
%     * signals(:,:) = mixmat*sources(:,:) + error
%

% Last update: 15-Sep-2015
% Author: Jun-ichiro Hirayama 

%Set default values for options, overwritten if given as inputs
p = struct('maxiter',10000,'pcafac',[],'complexmixing',false,'numcomps',[],...
           'deflation',false,'nonlintype','log','nonlinopts',1,...
           'display',true,'tolerance',1e-4);
p = parseargs(p, varargin);

if p.display
    fprintf('Performing complex-valued FastICA with these options:\n\n');
    disp(p)
    fprintf('\n')
end

%Set constants
[nSignal,nTime,nEpoch] = size(signals);
n = nTime .* nEpoch;
signals = signals(:,:);%concatenate all epochs
[f,g,h]=setnonlin(p.nonlintype,p.nonlinopts); %ICA nonlinearity

%Spatial whitening to make cov(X) unity; pseudo-cov may not
if p.complexmixing
    [X,whitemat,dewhitemat,sigavg]=whitening(signals,p.pcafac);
else
    [X,whitemat,dewhitemat,sigavg]=whitening(signals,p.pcafac,'real');
end
pCovX=(X*X.')./n; %for non-circularity correction

%Set the number of whitened signals and components
d = size(X,1);
if isempty(p.numcomps),  c = d;
else                     c = p.numcomps; end
%Randomly initialize orthogonal/unitary separation matrix 
W = randn(d,c);
if p.complexmixing, W=complex(W,randn(d,c)); end
W = projorth(W);
Wold = W;

%Fixed-point iteration of FastICA
Y = W'*X;
Y2 = abs(Y.*conj(Y));
if p.display
    frec=zeros(p.maxiter+1,1);
    frec(1)=sum(f(Y2(:)))./n;
end
for iter=1:p.maxiter
    gY2 = g(Y2);
    hY2 = h(Y2);
    %Fixed-point update with non-circularity correction
    % by Novey & Adali (2008)
    W = W*diag(mean(hY2.*Y2 + gY2,2)) - X*(gY2.*Y)'./n...
        + pCovX*conj(W)*diag(mean(hY2.*conj(Y).^2,2));
    %Original fixed-point iteration 
    % by Bingham & Hyvarinen (2000)
    %W = W*diag(mean(hY2.*Y2 + gY2,2)) - X*(gY2.*Y)'./n;
    if ~p.complexmixing, W = real(W); end
    W = projorth(W);
    Y = W'*X;
    Y2 = abs(Y.*conj(Y));
    if p.display
        frec(iter+1)=sum(f(Y2(:)))./n;
        plot(1:iter,frec(1:iter));drawnow;
    end
    %check if converged
    if norm(abs(W'*Wold)-eye(c),'fro') < c.*p.tolerance
        break;
    end
    Wold = W;
end
if iter==p.maxiter
    warning('Reached the maximum number of iterations.\n'); 
end
%Sort components according to the non-Gaussianity measure
fvals = sum(f(Y2),2)./n;
[fvals,order]=sort(fvals,'descend');
W=W(:,order);
Y=Y(order,:);
%Back to original space (non-whitened and non-centered)
mixmat = dewhitemat * W;
sepmat = W'*whitemat;
Y=bsxfun(@plus,Y,sepmat*sigavg);
sources = reshape(Y, [c,nTime,nEpoch]);

end

function W=projorth(W)
[E,D] = eig(W'*W);
W = W*E*diag(1./sqrt(diag(D)))*E';
end

function [f,g,h] = setnonlin(nonlintype,par)
%
% SETNONLIN Functions related to specific nonlinearity
%   f(|s|^2): objective 
%   g(|s|^2): gradient
%   h(|s|^2): hessian
%
if ~iscell(par), par = {par}; end
switch nonlintype
    case {'log','sqrt'}
      a = par{1};
end
eval(['f = @f_' nonlintype ';']);
eval(['g = @g_' nonlintype ';']);
eval(['h = @h_' nonlintype ';']);

%**function body ends here**

% Function definitions
%-----------------------------------------
% log
function f=f_log(q)
f = log(a + q);
end
function g=g_log(q)
g = 1./(a + q);
end
function h=h_log(q)
h = -1./(a + q).^2;
end

% kurt
function f=f_kurt(q)
f = 0.5.*(q.*q);
end
function g=g_kurt(q)
g = q;
end
function h=h_kurt(q)
h = ones(size(q));
end

% sqrt
function f=f_sqrt(q)
f = sqrt(a+q);
end
function g=g_sqrt(q)
g = 1./(2*sqrt(a + q));
end
function h=h_sqrt(q)
h = -1./(4.*((a + q).^(1.5)));
end
 
end