function [S,Pk,A,W,B,Eta,f]=lcmm(Xorig,K,varargin)
% LCMM Unified source separation an clustering based on latent
% coactivity mixture model
%
%  This is an implementation of the method presented in:
%   Unifying Blind Separation and Clustering for Resting-State EEG/MEG
%   Functional Connectivity Analysis.
%   Jun-ichiro Hirayama, Takeshi Ogawa & Aapo Hyvarinen
%   Neural Computation, Vol.27, No.7, pp.1373-1404, 2015.
%
%  This code requires minFunc toolbox available at
%    http://www.cs.ubc.ca/~schmidtm/Software/minFunc.html
%  
%  Usage:
%   >> [S,Pk,A,W,B,Eta,f]=lcmm(X,K,...)
%    
%  Input:
%     X: complex signals (after bandpass filtering and hilbert
%        transform). Each row is one channel's timeseries
%     K: number of clusters. (K=1,2,...)
%
%  Other input options may be given by name-value pairs, e.g.
%   >> [...] = lcmm(X,K,'maxiter',1000, 'verbose', false);
%
%  Available options:
%   'ftol'   : tolerance parameter for convergence judgement.
%   'maxiter': maximum number of quasi-newton iterations.
%   'verbose': if true, progress is displayed on the terminal.
%   'display': if true, paramters and objective values during
%               optimization are graphically displayed.
%   'pcafac' : if 1,2,.. dimensionality after PCA
%              if 0<fac<1, the fraction of variance explained.
%   'model'  : choice of prior model: currently either 'cauchy' or
%              'student2' can be used.
%   'W','B'  : if given, use them as initial parameters. If not given,
%              W is initialized by complex FastICA, and B at random.
%
%  Output:
%     S: Separated complex sources. Each row is a single source. 
%    Pk: Posterior probability of K states. Each row corresponds to
%        a single state. Use e.g. [~,idx] = max(Pk,[],1); to get cluster
%        index of each sample.
%     A: Real mixing matrix.   X = A*S + noise
%     W: Real demixing matrix. S = W*X
%     B: Matrix of coactivation (variance) patterns. B(:,k) is the
%        pattern of k-th state (cluster).
%   Eta: Estimated prior probability of the K states.
%     f: Log-likelihood of the learned model.

% Last update: 15-Sep-2015
% Author: Jun-ichiro Hirayama 

%% Process input options
opt = struct('ftol',1e-2,'maxiter',1000,'verbose',false,'display',false,...
             'pcafac',[],'model','cauchy','B',[],'W',[]);
opt = parseargs(opt,varargin);
if opt.verbose, display(opt); end

%% Data preprocessing
% whitening and dimensionality reduction by PCA
D_orig = size(Xorig,1);
[X,p.whiteMat,p.dewhiteMat,p.Xmean]=whitening(Xorig,opt.pcafac,'real');
[D,N] = size(X);
if opt.verbose
    fprintf('Dimensionality reduced %d -> %d\n', D_orig, D);
    fprintf('Sample size: %d \n', N);
    fprintf('Number of clusters: %d \n', K);
end
clear Xorig;

%% Estimation based on whitened data
p = lcmm_learn(X,K,p,opt);

% Post-processing 
S = p.W*X;
A = p.dewhiteMat/p.W;
W = p.W*p.whiteMat;
B = p.B;
Eta = exp(p.logPz);
Eta = Eta./sum(Eta(:));
Pk = p.Ez;
f = p.loglik;


% rescaling
Anorm = sqrt(sum(A.*A,1));
A     = A*diag(1./Anorm);
W     = diag(Anorm)*W;
S     = diag(Anorm)*S;
B     = diag(Anorm.*Anorm)*B;

% stdS = std(S,1,2);
% S    = scale_rows(S,1./stdS);
% A    = scale_cols(A,stdS);
% W    = scale_rows(W,1./stdS);
% B  = scale_rows(B,1./(stdS.*stdS));




end 
