function p=lcmm_learn(X,K,p,opt)
%Main routine for learning LCMM parameters with maximum likelihood

minopt =...
    struct('Method', 'lbfgs', 'MaxIter', opt.maxiter,'MaxFunEvals',opt.maxiter.*2,...
           'optTol', opt.ftol, 'useMex', 1, ...%'progTol'
           'DerivativeCheck','off','LS_init',0);
%   Some useful options for minFunc:
%     - 'DerivativeCheck': 'on'
%     - 'Corr': %when out of memory at 10-th iteration, try 9 or less

%minopt.Display='none';
if opt.verbose,  minopt.Display = 'iter';
else,  minopt.Display = 'none'; end

[D,N] = size(X);

%% Initialization by two-stage method

% initialize demixing matrix by complex fastica, or set to a given value
if isempty(opt.W)
    [~,~,p.W] = cfica(X,'display',false,'complexmixing',false);
    p.Wica=p.W;
else
    p.W = opt.W/p.whiteMat;
end
clear Xorig;

% initialize variance patterns
if isempty(opt.B)
    p.B = rand(D,K)+1e2; 
    p.B = (diag(1./sum(p.B,2))*p.B)./K;
else
    p.B = opt.B;
end

% prior weights for the K classes are set at uniform 
p.logPz = log(1./K).*ones(K,1);

%% Parameter estimation by a quasi-Newton method
[G,g,constNormalize] = lcmm_setnonlin(D,opt.model);

if opt.display
    minopt.outputFcn = @progfun;
    p.fprog = zeros(1,opt.maxiter);
end

%(optional) First learn the mixture model with fixed W
if false % set 'true' for this option
    initflag = true;
    v0=par2vec(p.W,p.B,p.logPz);
    [v,~,exitflag,p.minout]=minFunc(@objfun, v0, minopt);
    [p.W,p.B,p.logPz]=vec2par(v);
end

%Joint learning of two layers
initflag = false;
v0=par2vec(p.W,p.B,p.logPz);
[v,f,exitflag,p.minout]=minFunc(@objfun, v0, minopt);
[p.W,p.B,p.logPz]=vec2par(v);
p.frec = p.minout.trace.fval;
p.loglik = -f.*N;
p.iter = p.minout.iterations;
p.Ez = computeEz(p.W,p.B,p.logPz);

%=============================================================
% Nested functions
%=============================================================

%% Objective function and gradient
function [val,grad]=objfun(v)
    [W,B,logPz] = vec2par(v);
    Ez = computeEz(W,B,logPz);

    val=negloglik(W,B,logPz,Ez);

    % Compute gradient
    S   = W*X;
    S2  = realpow(abs(S),2);

    Nz = sum(Ez,2);
    P = g( (1./B)' * S2 );

    % gradient for B 
    gradB =  (1./B)*diag(Nz)./N - ((S2*(Ez.*P)')./(B.*B))./N;
    
    % gradient for logPz
    Pz = exp(logPz - repmat(logsumexp(logPz,1),K,1));
    gradLogPz = Pz - sum(Ez,2)./N;

    % gradient for W 
    if initflag
        gradW = zeros(size(W));
    else
        S_weighted = ((1./B) * (Ez.*P)).*S;
        gradW = ( S_weighted *X' + conj(S_weighted) *X.' )./N - 2.*pinv(W)'; 
    end
    grad = par2vec(gradW,gradB,gradLogPz);

end

% Nested function for conducting the E-step
function Ez=computeEz(W,B,logPz)
    logPz = logPz - repmat(logsumexp(logPz,1),K,1);
    % copmute posterior expectation
    Ez = repmat(logPz - sum(log(B)',2),1,N); %K-by-N
    Ez = - G((1./B)'*realpow(abs(W*X),2)) + Ez; %K-by-N
    Ez = exp(bsxfun(@minus, Ez, logsumexp(Ez,1)));
end

function f=negloglik(W,B,logPz,Ez)
%computation of negative log-likelihood
    logPz = logPz - repmat(logsumexp(logPz,1),K,1);
    eneg = bsxfun(@plus,sum(log(B)',2), G((1./B)'*realpow(abs(W*X),2)));
    eneg = bsxfun(@minus,eneg, logPz);
    f = -sum(logsumexp(-eneg,1),2)./N + constNormalize - 2.*log(abs(det(W)));
end

function v=par2vec(W,B,logPz)
    v = [W(:);B(:);logPz(:)];
end

function [W,B,logPz]=vec2par(vec)
    W = reshape(vec(1:(D.*D)),D,D);
    B = reshape(vec((D.*D) + (1:D.*K)),D,K);
    logPz = reshape(vec((D.*D+D.*K) + (1:K)),K,1);
end


function stop=progfun(x,iterationType,i,funEvals,f,t,gtd,g,d,optCond)
    [p.W,p.B,p.logPz] = vec2par(x);
    p.fprog(i+1)=f;
    p.iter = i+1;
    stop=0;
    p.Ez = computeEz(p.W,p.B,p.logPz);
    if opt.display
        subplot(1,3,1); imagesc(p.Ez);title('posterior'); drawnow
        subplot(1,3,2); bar(1:K, exp(p.logPz) );title('mixture ratio'); drawnow;
        subplot(1,3,3); plot(1:p.iter,p.fprog(1:p.iter),'m-'); title('objective'); drawnow;
        drawnow;
    end
end

end

